from .constraint import Constraint  # noqa
from .max_children import MaxChildren  # noqa
from .max_parents import MaxParents  # noqa
from .pin import Pin  # noqa
from .select_edge_nodes import SelectEdgeNodes  # noqa



from .appear import Appear  # noqa
from .costs import Costs  # noqa
from .edge_selection import EdgeSelection  # noqa
from .node_selection import NodeSelection  # noqa
from .split import Split  # noqa



from .constraints import SelectEdgeNodes
import logging
import numpy as np
import pylp

logger = logging.getLogger(__name__)


class Solver:
    """Create a solver for a given track graph.

    Args:

        track_graph (:class:`TrackGraph`):
            The graph of objects to track over time.

        skip_core_constraints (bool, default=False):
            If true, add no constraints to the solver at all. Otherwise, core
            constraints that ensure consistencies between selected nodes and
            edges are added.
    """

    def __init__(self, track_graph, skip_core_constraints=False):

        self.graph = track_graph
        self.variables = {}

        self.ilp_solver = None
        self.objective = None
        self.constraints = pylp.LinearConstraints()

        self.num_variables = 0
        self.costs = np.zeros((0,), dtype=np.float32)
        self.solution = None

        if not skip_core_constraints:
            self.add_constraints(SelectEdgeNodes())

    def add_costs(self, costs):

        logger.info("Adding %s costs...", type(costs).__name__)
        costs.apply(self)

    def add_constraints(self, constraints):

        logger.info("Adding %s constraints...", type(constraints).__name__)

        for constraint in constraints.instantiate(self):
            self.constraints.add(constraint)

    def solve(self, timeout=0.0, num_threads=1):

        self.objective = pylp.LinearObjective(self.num_variables)
        for i, c in enumerate(self.costs):
            logger.debug("Setting cost of var %d to %.3f", i, c)
            self.objective.set_coefficient(i, c)

        # TODO: support other variable types
        self.ilp_solver = pylp.LinearSolver(
            self.num_variables,
            pylp.VariableType.Binary,
            preference=pylp.Preference.Any)

        self.ilp_solver.set_objective(self.objective)
        self.ilp_solver.set_constraints(self.constraints)

        self.ilp_solver.set_num_threads(num_threads)
        if timeout > 0:
            self.ilp_solver.set_timeout(self.timeout)

        self.solution, message = self.ilp_solver.solve()
        if len(message):
            logger.info("ILP solver returned with: %s", message)

        return self.solution

    def get_variables(self, cls):

        if cls not in self.variables:
            self._add_variables(cls)

        return self.variables[cls]

    def _add_variables(self, cls):

        logger.info("Adding %s variables...", cls.__name__)

        keys = cls.instantiate(self)
        indices = self._allocate_variables(len(keys))
        variables = cls(self, {k: i for k, i in zip(keys, indices)})
        self.variables[cls] = variables

        for constraint in cls.instantiate_constraints(self):
            self.constraints.add(constraint)

    # TODO: add variable_type
    def _allocate_variables(self, num_variables):

        indices = range(
            self.num_variables,
            self.num_variables + num_variables
        )

        self.num_variables += num_variables
        self.costs.resize(self.num_variables, refcheck=False)

        return indices

    def add_variable_cost(self, index, cost):

        self.costs[index] += cost



from ..variables import NodeSplit
from .costs import Costs


class Split(Costs):

    def __init__(self, constant):

        self.constant = constant

    def apply(self, solver):

        split_indicators = solver.get_variables(NodeSplit)

        for index in split_indicators.values():
            solver.add_variable_cost(index, self.constant)



from .track_graph import TrackGraph
import networkx as nx


def get_tracks(graph, require_selected=False, selected_attribute='selected'):
    '''Get a generator of track graphs, each corresponding to one track
    (i.e., a connected component in the track graph).

    Args:

        require_selected (``bool``):

            If ``True``, consider only edges that have a selected_attribute
            attribute that is set to ``True``. Otherwise, each edge will be
            considered for the connected component analysis.

        selected_attribute (``str``):

            Only used if require_selected=True. Determines the attribute
            name to check if an edge is selected. Default value is
            'selected'.

    Returns:

        A generator object of graphs, one for each track.
    '''

    if require_selected:

        selected_edges = [
            e
            for e in graph.edges
            if (selected_attribute in graph.edges[e]
                and graph.edges[e][selected_attribute])
        ]
        graph = graph.edge_subgraph(selected_edges)

    if len(graph.nodes) == 0:
        return []

    return [
        TrackGraph(
            graph_data=graph.subgraph(g).copy(),
            frame_attribute=graph.frame_attribute,
            roi=graph.roi)
        for g in nx.weakly_connected_components(graph)
    ]



from ..variables import EdgeSelected
from .costs import Costs


class EdgeSelection(Costs):

    def __init__(self, weight, attribute='costs', constant=0.0):

        self.weight = weight
        self.attribute = attribute
        self.constant = constant

    def apply(self, solver):

        edge_variables = solver.get_variables(EdgeSelected)

        for edge, index in edge_variables.items():

            cost = (
                solver.graph.edges[edge][self.attribute] * self.weight +
                self.constant
            )

            solver.add_variable_cost(index, cost)



from ..variables import NodeSelected, EdgeSelected
from .constraint import Constraint
import pylp


class Pin(Constraint):
    """Enforces the selection of certain nodes and edges based on the value of
    a given attribute.

    Every node or edge that has the given attribute will be selected if the
    attribute value is ``True`` (and not selected if the attribute value is
    ``False``). The solver will only determine the selection of nodes and edges
    that do not have this attribute.

    This constraint is useful to complete partial tracking solutions, e.g.,
    when users have already manually selected (or unselected) certain nodes or
    edges.

    Args:

        attribute (string):
            The name of the node/edge attribute to use.
    """

    def __init__(self, attribute):

        self.attribute = attribute

    def instantiate(self, solver):

        node_indicators = solver.get_variables(NodeSelected)
        edge_indicators = solver.get_variables(EdgeSelected)

        must_select = [
            node_indicators[node]
            for node, value in solver.graph.nodes.data(self.attribute)
            if value is True
        ] + [
            edge_indicators[(u, v)]
            for u, v, value in solver.graph.edges.data(self.attribute)
            if value is True
        ]

        must_not_select = [
            node_indicators[node]
            for node, value in solver.graph.nodes.data(self.attribute)
            if value is False
        ] + [
            edge_indicators[(u, v)]
            for u, v, value in solver.graph.edges.data(self.attribute)
            if value is False
        ]

        must_select_constraint = pylp.LinearConstraint()
        must_not_select_constraint = pylp.LinearConstraint()

        for index in must_select:
            must_select_constraint.set_coefficient(index, 1)
        for index in must_not_select:
            must_not_select_constraint.set_coefficient(index, 1)

        must_select_constraint.set_relation(pylp.Equal)
        must_not_select_constraint.set_relation(pylp.Equal)

        must_select_constraint.set_value(len(must_select))
        must_not_select_constraint.set_value(0)

        return [must_select_constraint, must_not_select_constraint]



from ..variables import NodeSelected, EdgeSelected
from .constraint import Constraint
import pylp


class SelectEdgeNodes(Constraint):
    """Ensures that if an edge (u, v) is selected, u and v have to be selected
    as well.

    Adds the following linear constraint for each edge::

      2 * edge(u, v) - u - v <= 0

    This constraint will be added by default to any :class:`Solver` instance.
    """

    def instantiate(self, solver):

        node_indicators = solver.get_variables(NodeSelected)
        edge_indicators = solver.get_variables(EdgeSelected)

        constraints = []
        for edge in solver.graph.edges:

            u, v = edge

            ind_e = edge_indicators[edge]
            ind_u = node_indicators[u]
            ind_v = node_indicators[v]

            constraint = pylp.LinearConstraint()
            constraint.set_coefficient(ind_e, 2)
            constraint.set_coefficient(ind_u, -1)
            constraint.set_coefficient(ind_v, -1)
            constraint.set_relation(pylp.Relation.LessEqual)
            constraint.set_value(0)
            constraints.append(constraint)

        return constraints



from .edge_selected import EdgeSelected  # noqa
from .node_appear import NodeAppear  # noqa
from .node_selected import NodeSelected  # noqa
from .node_split import NodeSplit  # noqa
from .variable import Variable  # noqa



from .edge_selected import EdgeSelected
from .node_selected import NodeSelected
from .variable import Variable
import pylp


class NodeAppear(Variable):

    @staticmethod
    def instantiate(solver):
        return solver.graph.nodes

    @staticmethod
    def instantiate_constraints(solver):

        appear_indicators = solver.get_variables(NodeAppear)
        node_indicators = solver.get_variables(NodeSelected)
        edge_indicators = solver.get_variables(EdgeSelected)

        constraints = []
        for node in solver.graph.nodes:

            prev_edges = solver.graph.prev_edges(node)
            num_prev_edges = len(prev_edges)

            # Ensure that the following holds:
            #
            # appear = 1 <=> sum(prev_selected) = 0 and selected
            # appear = 0 <=> sum(prev_selected) > 0 or not selected
            #
            # Two linear constraints are needed for that:
            #
            # let s = num_prev * selected - sum(prev_selected)
            # (1) s - appear <= num_prev - 1
            # (2) s - appear * num_prev >= 0

            constraint1 = pylp.LinearConstraint()
            constraint2 = pylp.LinearConstraint()

            # set s for both constraints:

            # num_prev * selected
            constraint1.set_coefficient(
                node_indicators[node],
                num_prev_edges)
            constraint2.set_coefficient(
                node_indicators[node],
                num_prev_edges)

            # - sum(prev_selected)
            for prev_edge in prev_edges:
                constraint1.set_coefficient(
                    edge_indicators[prev_edge],
                    -1.0)
                constraint2.set_coefficient(
                    edge_indicators[prev_edge],
                    -1.0)

            # constraint specific parts:

            # - appear
            constraint1.set_coefficient(
                appear_indicators[node],
                -1.0)

            # - appear * num_prev
            constraint2.set_coefficient(
                appear_indicators[node],
                -num_prev_edges)

            constraint1.set_relation(pylp.Relation.LessEqual)
            constraint2.set_relation(pylp.Relation.GreaterEqual)

            constraint1.set_value(num_prev_edges - 1)
            constraint2.set_value(0)

            constraints.append(constraint1)
            constraints.append(constraint2)

        return constraints



from ..variables import EdgeSelected
from .constraint import Constraint
import pylp


class MaxChildren(Constraint):
    """Ensures that every selected node has no more than ``max_children``
    selected edges to the next frame.

    Adds the following linear constraint for each node::

      sum(next) <= max_children

    Args:

        max_children (int):
            The maximum number of children allowed.
    """

    def __init__(self, max_children):

        self.max_children = max_children

    def instantiate(self, solver):

        edge_indicators = solver.get_variables(EdgeSelected)

        constraints = []
        for node in solver.graph.nodes:

            constraint = pylp.LinearConstraint()

            # all outgoing edges
            for edge in solver.graph.next_edges(node):
                constraint.set_coefficient(edge_indicators[edge], 1)

            # relation, value
            constraint.set_relation(pylp.Relation.LessEqual)

            constraint.set_value(self.max_children)
            constraints.append(constraint)

        return constraints



from abc import ABC, abstractmethod


class Constraint(ABC):

    @abstractmethod
    def instantiate(self, solver):
        """Create and return specific linear constraints for the given solver.

        Args:

            solver (:class:`Solver`):
                The solver instance to create linear constraints for.

        Returns:

            A list of :class:`pylp.LinearConstraint`.
        """
        pass



from .edge_selected import EdgeSelected
from .variable import Variable
import pylp


class NodeSplit(Variable):

    @staticmethod
    def instantiate(solver):
        return solver.graph.nodes

    @staticmethod
    def instantiate_constraints(solver):

        split_indicators = solver.get_variables(NodeSplit)
        edge_indicators = solver.get_variables(EdgeSelected)

        constraints = []
        for node in solver.graph.nodes:

            next_edges = solver.graph.next_edges(node)

            # Ensure that the following holds:
            #
            # split = 0 <=> sum(next_selected) <= 1
            # split = 1 <=> sum(next_selected) > 1
            #
            # Two linear constraints are needed for that:
            #
            # (1) 2 * split - sum(next_selected) <= 0
            # (2) (num_next - 1) * split - sum(next_selected) >= -1

            constraint1 = pylp.LinearConstraint()
            constraint2 = pylp.LinearConstraint()

            constraint1.set_coefficient(
                split_indicators[node],
                2.0)
            constraint2.set_coefficient(
                split_indicators[node],
                len(next_edges) - 1.0)

            for next_edge in next_edges:
                constraint1.set_coefficient(
                    edge_indicators[next_edge],
                    -1.0)
                constraint2.set_coefficient(
                    edge_indicators[next_edge],
                    -1.0)

            constraint1.set_relation(pylp.Relation.LessEqual)
            constraint2.set_relation(pylp.Relation.GreaterEqual)

            constraint1.set_value(0.0)
            constraint2.set_value(-1.0)

            constraints.append(constraint1)
            constraints.append(constraint2)

        return constraints



from abc import ABC, abstractmethod


class Costs(ABC):

    @abstractmethod
    def apply(self, solver):
        """Apply costs to the given solver. Use `solver.get_variables` and
        `solver.add_variable_cost`.

        Args:

            solver (:class:`Solver`):
                The solver to create costs for.
        """
        pass



from ..variables import NodeAppear
from .costs import Costs


class Appear(Costs):

    def __init__(self, constant):

        self.constant = constant

    def apply(self, solver):

        appear_indicators = solver.get_variables(NodeAppear)

        for index in appear_indicators.values():
            solver.add_variable_cost(index, self.constant)



from setuptools import setup

setup(
        name='motile',
        version='0.1.1',
        description='Multi-Object Tracker using Integer Linear Equations',
        url='https://github.com/funkelab/motile',
        author='Jan Funke',
        author_email='funkej@janelia.hhmi.org',
        license='MIT',
        packages=[
            'motile',
            'motile.constraints',
            'motile.costs',
            'motile.variables'
        ],
        install_requires=[
            'networkx',
            'pylp',
            'numpy'
        ]
)



import logging
import networkx as nx

logger = logging.getLogger(__name__)


class TrackGraph(nx.DiGraph):
    """A graph of objects with positions in time and space, and inter-frame
    edges between them.

    Args:

    graph_data (optional):

        Optional graph data to pass to the networkx.Graph constructor as
        ``incoming_graph_data``. This can be used to populate a track graph
        with entries from a generic networkx graph.

    frame_attribute (``string``, optional):

        The name of the node attribute that corresponds to the frame of the
        object. Defaults to "t".
    """

    def __init__(
            self,
            graph_data=None,
            frame_attribute='t'):

        super().__init__(incoming_graph_data=graph_data)

        self.frame_attribute = frame_attribute
        self._graph_changed = True

        self._update_metadata()

    def prev_edges(self, node):
        '''Get all edges that point forward into ``node``.'''

        return self.in_edges(node)

    def next_edges(self, node):
        '''Get all edges that point forward out of ``node``.'''

        return self.out_edges(node)

    def get_frames(self):
        '''Get a tuple ``(t_begin, t_end)`` of the first and last frame
        (exclusive) this track graph has nodes for.'''

        self._update_metadata()

        return (self.t_begin, self.t_end)

    def nodes_by_frame(self, t):
        '''Get all nodes in frame ``t``.'''

        self._update_metadata()

        if t not in self._nodes_by_frame:
            return []
        return self._nodes_by_frame[t]

    def _update_metadata(self):

        if not self._graph_changed:
            return

        self._graph_changed = False

        if self.number_of_nodes() == 0:

            self._nodes_by_frame = {}
            self.t_begin = None
            self.t_end = None
            return

        self._nodes_by_frame = {}
        for node, data in self.nodes(data=True):
            t = data[self.frame_attribute]
            if t not in self._nodes_by_frame:
                self._nodes_by_frame[t] = []
            self._nodes_by_frame[t].append(node)

        frames = self._nodes_by_frame.keys()
        self.t_begin = min(frames)
        self.t_end = max(frames) + 1

        # ensure edges point forwards in time
        for u, v in self.edges:
            t_u = self.nodes[u][self.frame_attribute]
            t_v = self.nodes[v][self.frame_attribute]
            assert t_u < t_v, \
                f"Edge ({u}, {v}) does not point forwards in time, but from " \
                f"frame {t_u} to {t_v}"

        self._graph_changed = False

    # wrappers around node/edge add/remove methods:

    def add_node(self, n, **attr):
        super().add_node(n, **attr)
        self._graph_changed = True

    def add_nodes_from(self, nodes, **attr):
        super().add_nodes_from(nodes, **attr)
        self._graph_changed = True

    def remove_node(self, n):
        super().remove_node(n)
        self._graph_changed = True

    def remove_nodes_from(self, nodes):
        super().remove_nodes_from(nodes)
        self._graph_changed = True

    def add_edge(self, u, v, **attr):
        super().add_edge(u, v, **attr)
        self._graph_changed = True

    def add_edges_from(self, ebunch_to_add, **attr):
        super().add_edges_from(ebunch_to_add, **attr)
        self._graph_changed = True

    def add_weighted_edges_From(self, ebunch_to_add):
        super().add_weighted_edges_From(ebunch_to_add)
        self._graph_changed = True

    def remove_edge(self, u, v):
        super().remove_edge(u, v)
        self._graph_changed = True

    def update(self, edges, nodes):
        super().update(edges, nodes)
        self._graph_changed = True

    def clear(self):
        super().clear()
        self._graph_changed = True

    def clear_edges(self):
        super().clear_edges()
        self._graph_changed = True



from ..variables import NodeSelected
from .costs import Costs


class NodeSelection(Costs):

    def __init__(self, weight, attribute='costs', constant=0.0):

        self.weight = weight
        self.attribute = attribute
        self.constant = constant

    def apply(self, solver):

        node_variables = solver.get_variables(NodeSelected)

        for node, index in node_variables.items():

            cost = (
                solver.graph.nodes[node][self.attribute] * self.weight +
                self.constant
            )

            solver.add_variable_cost(index, cost)



from abc import ABC, abstractmethod


class Variable(ABC):

    @staticmethod
    @abstractmethod
    def instantiate(solver):
        """Create and return variable keys for the given solver.

        Args:

            solver (:class:`Solver`):
                The solver instance to create variables for.

        Returns:

            A list of keys (anything that is hashable, e.g., nodes of a graph),
            one for each variable to create.
        """
        pass

    @staticmethod
    def instantiate_constraints(solver):
        return []

    def __init__(self, solver, index_map):
        self._solver = solver
        self._index_map = index_map

    def __repr__(self):

        rs = []
        for key, index in self._index_map.items():
            r = type(self).__name__ + f"({key}): "
            r += f"cost={self._solver.costs[index]} "
            if self._solver.solution is not None:
                r += f"value={self._solver.solution[index]}"
            else:
                r += "value=None"
            rs.append(r)
        return "\n".join(rs)

    def __getitem__(self, key):
        return self._index_map[key]

    def items(self):
        return self._index_map.items()

    def keys(self):
        return self._index_map.keys()

    def values(self):
        return self._index_map.values()
